;;;; -*- mode: lisp -*-
;;;;
;;;; $Id: base64.lisp,v 1.3 2005/02/07 17:45:41 scaekenberghe Exp $
;;;;
;;;; This is a minimal standalone Common Lisp HTTP Server
;;;;
;;;; Copyright (C) 2005 Sven Van Caekenberghe, Beta Nine BVBA.
;;;;
;;;; You are granted the rights to distribute and use this software
;;;; as governed by the terms of the Lisp Lesser General Public License
;;;; (http://opensource.franz.com/preamble.html), also known as the LLGPL.

(in-package :s-http-server)

;; globals

(defvar *http-server-identification* 
  (format nil "S-HTTP-SERVER ~a ~a" (lisp-implementation-type) (lisp-implementation-version))
  "Identification string sent as value of the 'Server' HTTP Response Header")

(defvar *http-server-port* 1701
  "Default port used when creating a new S-HTTP-SERVER")

(defclass s-http-server ()
  ((port :reader get-port :initarg :port :initform *http-server-port*)
   (name :accessor get-name :initarg :name :initform "s-http-server")
   (debug-mode :accessor get-debug-mode :initarg :debug-mode :initform t)
   (server-process :accessor get-server-process :initform nil)
   (log-stream :accessor get-log-stream :initarg :log-stream :initform nil)
   (contexts :accessor get-contexts :initarg :contexts 
             :initform '((s-http-server-handler "/s-http-server" :builtin))))
  (:documentation "The object representing a minimal standalone HTTP Server"))

(setf (documentation 'get-port 'function)
      "Get the TCP port used by this S-HTTP-SERVER"
      (documentation 'get-name 'function)
      "Get the current name of this S-HTTP-SERVER"
      (documentation 'get-debug-mode 'function)
      "Get the current mode of debugging of this S-HTTP-SERVER, t is on, nil is off"
      (documentation 'get-server-process 'function)
      "Get the current server process used by this S-HTTP-SERVER, nil if not running"
      (documentation 'get-log-stream 'function)
      "Get the current stream used by this S-HTTP-SERVER for logging, nil means no logging"
      (documentation 'get-contexts 'function)
      "Get the current list of context bindings used by this S-HTTP-SERVER")

#-allegro
(setf (documentation '(setf get-name) 'function)
      "Set the name of this S-HTTP-SERVER"
      (documentation '(setf get-debug-mode) 'function)
      "Set the current debugging mode of this S-HTTP-SERVER, t is on, nil is off"
      (documentation '(setf get-log-stream) 'function)
      "Set the stream this S-HTTP-SERVER uses for logging, nil means no logging")

(defun make-s-http-server (&key 
                           (port *http-server-port*)
                           (name "s-http-server")
                           (log-stream *standard-output*))
  "Create a new object representing an S-HTTP-SERVER"
  (make-instance 's-http-server
                 :port port
                 :name name
                 :log-stream log-stream))

(defmethod print-object ((s-http-server s-http-server) stream)
  (print-unreadable-object (s-http-server stream :type t :identity t)
    (with-slots (name port server-process)
        s-http-server
      (format stream "~s port ~d ~a" name port (if server-process "running" "not running")))))

(defclass http-request ()
  ((method :accessor get-method :initarg :method :initform :GET)
   (uri :accessor get-uri :initarg :uri :initform (puri:parse-uri "/"))
   (http-version :accessor get-http-version :initarg :http-version :initform "HTTP/1.1")
   (headers :accessor get-headers :initarg :headers :initform '())
   (keep-alive :accessor get-keep-alive :initarg :keep-alive :initform t))
  (:documentation "The object representing an HTTP request as being handled by the S-HTTP-SERVER"))

(setf (documentation 'get-method 'function)
      "Get the method (keyword :get :put :post :delete ..) of this HTTP request"
      (documentation 'get-uri 'function)
      "Get the URI object of this HTTP request"
      (documentation 'get-http-version 'function)
      "Get the HTTP version string of this HTTP request"
      (documentation 'get-headers 'function)
      "Get the dotted alist (:keyword . 'value') of request headers of this HTTP request"
      (documentation 'get-keep-alive 'function)
      "Is this a keep-alive request (either 1.0 or 1.1)")

(defgeneric get-path (http-request)
  (:method ((http-request http-request))
   (puri:uri-path (get-uri http-request)))
  (:documentation "Get the path of this HTTP request"))

(defmethod print-object ((http-request http-request) stream)
  (print-unreadable-object (http-request stream :type t :identity t)
    (format stream "~a ~s" (get-method http-request) (get-path http-request))))

;; generics

(defgeneric start-server (server)
  (:documentation "Start the server"))

(defgeneric stop-server (server)
  (:documentation "Stop the server"))

(defgeneric logm (server format-string &rest args)
  (:documentation "Log a formatted message"))

(defgeneric handle-http-server-connection (server connection-id client-socket-stream)
  (:documentation "Handle a new connection request in a new process"))

(defgeneric find-handler (server http-request)
  (:documentation "Given http-request select a handler from server"))

(defgeneric register-context-handler (server context-prefix handler-function &key arguments at-end-p do-not-replace-p)
  ;; optional handler arguments can be specified
  ;; normally, an existing context binding with the same prefix is overwritten
  ;; normally, new handlers are pushed at the front of the context bindings list
  ;; if at-end-p is t, a new binding will be added at the end of the context bindings list
  ;; if do-not-replace-p is t, an existing binding will not be overwritten and a new one will be created
  (:documentation "Configure server so that every request starting with context-prefix is sent to handler-function"))

(defgeneric unregister-context-handler (server context-prefix &key only-first-p only-last-p)
  ;; normally, all context bindings matching exactly the specified prefix are deleted
  ;; if only-first-p is t, only the first context binding with prefix is deleted
  ;; if only-last-p is t , only the last context binding with prefix is deleted
  ;; if both only-fast-p and only-last-p are t, an error is signalled
  (:documentation "Remove any configuration of server for context-prefix"))

;; setup

(defmethod start-server ((s-http-server s-http-server))
  (stop-server s-http-server)
  (let ((connection-id 0))
    (flet ((connection-handler (client-socket-stream)
             (incf connection-id)
             (s-sysdeps:run-process (format nil "connection-handler-~d" connection-id)
                                    #'handle-http-server-connection 
                                    s-http-server
                                    connection-id
                                    client-socket-stream)))
      (let ((process (s-sysdeps:start-standard-server :port (get-port s-http-server)
                                                      :name (get-name s-http-server)
                                                      :connection-handler #'connection-handler)))
        (setf (get-server-process s-http-server) process)
        (logm s-http-server "Started a new server on port ~d" (get-port s-http-server)))))
  s-http-server)

(defmethod stop-server ((s-http-server s-http-server))
  (let ((process (get-server-process s-http-server)))
    (when process
      (s-sysdeps:kill-process process)
      (setf (get-server-process s-http-server) nil)
      (logm s-http-server "Stopped server")))
  s-http-server)

(defmethod register-context-handler ((s-http-server s-http-server) context-prefix handler-function 
                                     &key arguments at-end-p do-not-replace-p)
  (let* ((new-handler-binding `(,handler-function ,context-prefix ,@arguments))
         (context-bindings (get-contexts s-http-server))
         (existing-binding (find context-prefix context-bindings :key #'second :test #'string=)))
    (if (or do-not-replace-p (null existing-binding))
        (if at-end-p
            (setf (get-contexts s-http-server) (append context-bindings (list new-handler-binding)))
          (push new-handler-binding (get-contexts s-http-server)))
      (loop :for binding :in context-bindings :do
            (destructuring-bind (function prefix &rest args)
                binding
              (declare (ignore function args))
              (when (string= prefix context-prefix)
                (setf (car binding) handler-function
                      (cddr binding) arguments)
                ;; we assume there is only one binding, more doesn't really make sense
                (return binding)))))))

(defmethod unregister-context-handler ((s-http-server s-http-server) context-prefix 
                                       &key only-first-p only-last-p)
  (let ((context-bindings (get-contexts s-http-server)))
    (cond ((and only-first-p only-last-p) (error "You cannot specify both only-first-p and only-last-p"))
          (only-first-p 
           (setf (get-contexts s-http-server) (delete context-prefix context-bindings
                                                      :key #'second :test #'string= :count 1)))
          (only-last-p 
           (setf (get-contexts s-http-server) (delete context-prefix context-prefix 
                                                      :key #'second :test #'string= :count 1 :from-end t)))
          (t 
           (setf (get-contexts s-http-server) (delete context-prefix context-bindings
                                                      :key #'second :test #'string= :count 1 :from-end t))))))

(defmethod logm ((s-http-server s-http-server) format-string &rest args)
  (let ((out (get-log-stream s-http-server)))
    (when out
      (format out ";; ~a: " (string-upcase (get-name s-http-server)))
      (apply #'format out format-string args)
      (terpri out))))

;; low level input/output

(defun read-http-request-line (stream)
  (let ((line (read-line stream nil)))
    (when line
      (string-trim '(#\return #\newline) line))))

(defun write-http-response-line (string &optional (stream *standard-output*))
  "Write string to stream, ending with the HTTP end of line convention (CR+LF)"
  (write-string string stream)
  (write-char #\return stream)
  (write-char #\linefeed stream))

(defun format-http-response-line (stream format-string &rest args)
  (write-http-response-line (apply #'format nil format-string args) stream))

;; parsing requests

(define-condition missing-http-request-line (error) ())

(defun parse-http-request-line (stream)
  (let* ((line (read-http-request-line stream))
         (tokens (s-utils:tokens line :separators '(#\Space))))
    (if tokens
        (values (intern (first tokens) :keyword)
                (puri:parse-uri (second tokens))
                (third tokens))
      (error 'missing-http-request-line))))

(defun parse-http-request-headers (stream)
  (loop :for line = (read-http-request-line stream)
        :until (or (null line)
                   (zerop (length line))
                   (char= (elt line 0) #\return)
                   (char= (elt line 0) #\linefeed))
        :collect (let ((colon (position #\: line)))
                   (cons (intern (string-upcase (subseq line 0 colon)) :keyword)
                         (string-trim '(#\space) (subseq line (1+ colon)))))))

(defun parse-http-request (stream)
  (multiple-value-bind (http-method uri http-version)
      (parse-http-request-line stream)
    (let* ((request-headers (parse-http-request-headers stream))
           (http-request (make-instance 'http-request 
                                        :method http-method
                                        :uri uri
                                        :http-version http-version
                                        :headers request-headers)))
      (when (and (string-equal http-version "HTTP/1.0")
                 (string-not-equal (cdr (assoc "Connection" request-headers :test #'string-equal)) 
                                   "Keep-Alive"))
        (setf (get-keep-alive http-request) nil))
      http-request)))

;; write and generating responses

(defun write-http-response-status-line (stream &optional (status-code 200) (string "OK") (http-version "HTTP/1.1"))
  "Write an HTTP Response Status line to stream, using status-code string and http-version"
  (format-http-response-line stream "~a ~d ~a" http-version status-code string))

(defun write-http-response-headers (headers stream)
  "Write the headers alist as HTTP Response Headers to stream"
  (loop :for (header-key . header-value) :in headers
        :do (format-http-response-line stream "~a: ~a" header-key header-value)))

(defun response-date (&optional (universal-time (get-universal-time)))
  "Generate a GMT HTTP Response Date"
  (s-utils:format-universal-time universal-time
                                 :format '(:day-name ", " :date2 #\Space :month-name #\Space :year #\Space 
                                           :hour #\: :minute #\: :second " GMT")
                                 :timezone 0))

(defun standard-http-response-headers (http-request &key (content-type "text/plain") content-length)
  "Generate the standard headers alist given context-type and context-length, managing old-style Keep-Alive"
  `(("Server" . ,*http-server-identification*)
    ("Date" . ,(response-date))
    ("Content-Type" . ,content-type)
    ,@(when content-length
        `(("Content-Length" . ,content-length)))
    ,@(when (and (get-keep-alive http-request) 
                 (string-equal (get-http-version http-request) "HTTP/1.0"))
        `(("Connection" . "Keep-Alive")))))

(defun escape (string)
  (with-output-to-string (stream)
    (loop :for char :across string
          :do (case char
                (#\& (write-string "&amp;" stream))
                (#\< (write-string "&lt;" stream))
                (#\> (write-string "&gt;" stream))
                (#\" (write-string "&quot;" stream))
                ((#\newline #\return #\tab) (write-char char stream))
                (t (if (and (<= 32 (char-code char))
                            (<= (char-code char) 126))
                       (write-char char stream)
                     (progn
                       (write-string "&#x" stream)
                       (write (char-code char) :stream stream :base 16)
                       (write-char #\; stream))))))))

(defvar *doctype-html-401-transitional*
  "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">")

(defvar *doctype-html-401-strict*
  "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">")

(defun standard-http-html-message-response (http-request stream title message &optional (status 200) (string "OK"))
  "Generate and write a standard HTML message as HTTP Response using title, message, status and string"
  (let ((content (with-output-to-string (out)
                   (format out 
                           "~a~%<html lang=\"en\"><head><title>~a</title></head><body><h1>~a</h1>~a</body></html>"
                           *doctype-html-401-strict* title title message))))
    (write-http-response-status-line stream status string)
    (write-http-response-headers (standard-http-response-headers http-request
                                                                 :content-type "text/html"
                                                                 :content-length (length content)) 
                                 stream)
    (write-http-response-line "" stream)
    (write-string content stream)))

(defun standard-http-html-error-response (http-request stream code reason extra)
  "Generate and write a standard HTML error as HTTP Response using code, reason and extra"
  (standard-http-html-message-response http-request 
                                       stream
                                       reason
                                       (format nil "<p>~d - ~a: ~a</p>" code reason (escape (prin1-to-string extra)))
                                       code
                                       reason))

;; core server implementation (http request/repsonse loop and dispatching to handlers)

(defmethod find-handler ((s-http-server s-http-server) http-request)
  (let ((path (get-path http-request)))
    (loop :for context-binding :in (get-contexts s-http-server)
          :do (destructuring-bind (handler context &rest rest)
                  context-binding
                (declare (ignore handler rest))
                (if (string= path context :end1 (min (length context) (length path)))
                    (return-from find-handler context-binding))))))

(defmethod handle-one-http-request-response ((s-http-server s-http-server) connection-id stream)
  (let* ((http-request (parse-http-request stream))
         (handler (find-handler s-http-server http-request)))
    (logm s-http-server "[~d] Handling ~s" connection-id http-request) 
    (if handler
        (if (get-debug-mode s-http-server)
            (funcall (first handler) 
                     s-http-server (rest handler) http-request stream)
          (multiple-value-bind (result condition)
              (ignore-errors (funcall (first handler) 
                                      s-http-server (rest handler) http-request stream))
            (unless result 
              (logm s-http-server "[~d] Handler ~s failed for ~s" connection-id handler http-request)
              (standard-http-html-error-response http-request stream 500 "Internal Server Error" condition))))
      (progn
        (logm s-http-server "[~d] No handler found for ~s" connection-id http-request)
        (standard-http-html-error-response http-request stream 404 "Resource Not Found" (get-path http-request))))
    (finish-output stream)
    http-request))

(defmethod handle-http-server-connection ((s-http-server s-http-server) connection-id stream)
  (loop
   (if (get-debug-mode s-http-server)
       (unless (handler-case
                   (get-keep-alive (handle-one-http-request-response s-http-server connection-id stream))
                 (missing-http-request-line () nil))
         (return))
     (unless (ignore-errors (get-keep-alive (handle-one-http-request-response s-http-server connection-id stream)))
       (return))))
  (close stream))

;; an S-HTTP-HANDLER is a function taking 4 arguments:
;; - the s-http-server object representing the server itself
;; - the handler binding (a list whose first element is always the context that matched)
;; - the current http-request object (containing request method, uri, version and the headers alist)
;; - the stream to the client (where the request line and headers are already consumed)
;; the handler should handle the request, outputing 3 things (in order):
;; - a response status line
;; - the response headers
;; - the contents
;; finally, the handler should return t when it was successful

;; the builtin handler

(defun s-http-server-handler (s-http-server handler http-request stream)
  "The builtin S-HTTP-SERVER testing/debugging handler returning a simple status/echo/snoop page"
  (logm s-http-server "Running builtin s-http-server handler")
  (let ((body (with-output-to-string (out)
                (format out "<p>Welcome to ~a</p>" *http-server-identification*)
                (format out "<p>This is ~s running on port ~d</p>" (get-name s-http-server) (get-port s-http-server))
                (format out "<p>S-HTTP-SERVER-HANDLER handling a ~a request for path ~s and version ~a</p>" 
                        (get-method http-request) (get-path http-request) (get-http-version http-request))
                (format out "<p>Remote host is ~s</p>" (s-sysdeps:get-socket-stream-property stream :remote-host))
                (format out "<p>Current thread is ~s</p>" (escape (prin1-to-string (s-sysdeps:current-process))))
                (format out "<p>Active handler binding is ~s</p>" handler)
                (format out "<p>Contexts:</p><table border='1' width='100%'>")
                (format out "<tr><th>Handler</th><th>Context</th><th>Parameters</th></tr>")
                (loop :for b :in (get-contexts s-http-server) 
                      :do (format out "<tr><td>~a</td><td>~a</td><td>~a</td></tr>" 
                                  (escape (prin1-to-string (first b))) 
                                  (second b) 
                                  (escape (prin1-to-string (rest (rest b))))))
                (format out "</table>")
                (format out "</ul><p>Request Headers:</p>")
                (format out "<table border='1' width='100%'><tr><th>Key</th><th>Value</th></tr>")
                (loop :for (k . v) :in (get-headers http-request) 
                      :do (format out "<tr><td>~a</td><td>~a</td></tr>" k v))
                (format out "</table>"))))
    (standard-http-html-message-response http-request stream "S-HTTP-SERVER" body)
    t))

;; the static resource (file server) handler

(defconstant +basic-mime-type-suffix-map+ 
  '(("html" . "text/html") ("htm" . "text/html") ("txt" . "text/plain")
    ("gif" . "image/gif") ("jpg" . "image/jpeg") ("jpeg" . "image/jpg") ("png" . "image/png")))

(defconstant +known-mime.type-locations+ '("/etc/httpd/mime.types" "/etc/mime.types"))

(defparameter *mime-type-suffix-map* 
  (let ((map (make-hash-table :test 'equal)))
    (loop :for (suffix . mime-type) :in +basic-mime-type-suffix-map+ 
          :do (setf (gethash suffix map) mime-type))
    (loop :for location :in +known-mime.type-locations+
          :do (when (probe-file location)
                (with-open-file (in location)
                  (loop :for line = (read-line in nil)
                        :until (null line)
                        :unless (or (zerop (length line))
                                    (member (elt line 0) '(#\return #\linefeed #\#)))
                        :do (let* ((tokens (s-utils:tokens line :separators '(#\tab)))
                                   (mime-type (string-trim '(#\space) (first tokens)))
                                   (suffixes (s-utils:tokens (second tokens) :separators '(#\space))))
                              (loop :for suffix :in suffixes
                                    :do (setf (gethash suffix map) mime-type)))))))
    map))

(defun mime-type-for-pathname (pathname)
  (or (gethash (string-downcase (or (pathname-type pathname) "")) *mime-type-suffix-map*)
      "application/octet-stream"))

(defun compute-real-resource-pathname (root path context)
  (let* ((real-root (pathname root))
         (real-root-dir-components (rest (pathname-directory real-root)))
         (context-pathname (pathname (concatenate 'string context "/")))
         (context-components (rest (pathname-directory context-pathname))) 
         (uri-pathname (pathname path))
         (uri-dir-components (rest (pathname-directory uri-pathname)))
         (difference (mismatch context-components uri-dir-components :test #'string=)))
    (setf uri-dir-components (when difference (subseq uri-dir-components difference)))
    (make-pathname :name (or (pathname-name uri-pathname) "index")
                   :type (or (pathname-type uri-pathname) "html")
                   :directory `(:absolute 
                                ,@real-root-dir-components
                                ,@uri-dir-components))))

(defun static-resource-handler (s-http-server handler http-request stream)
  "Host static resources from a document root"
  (destructuring-bind (context document-root)
      handler
    (let* ((path (get-path http-request))
           (resource-pathname (compute-real-resource-pathname document-root path context)))
      (if (probe-file resource-pathname)
          (let ((mime-type (mime-type-for-pathname resource-pathname)))
            (logm s-http-server "Serving ~s" resource-pathname)
            (with-open-file (in resource-pathname :element-type '(unsigned-byte 8))
              (write-http-response-status-line stream)
              (write-http-response-headers (standard-http-response-headers http-request
                                                                           :content-type mime-type 
                                                                           :content-length (file-length in))
                                           stream)
              (write-http-response-headers `(("Last-Modified" . ,(response-date (file-write-date in))))
                                           stream)
              (write-http-response-line "" stream)
              (s-utils:copy-stream in stream (make-array 4096 :element-type '(unsigned-byte 8)))))
        (progn
          (logm s-http-server "Failed to find ~s" resource-pathname)
          (standard-http-html-error-response http-request stream 404 "Resource Not Found" path)))
      t)))

;;;; eof
